#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include "sqlite.h"
#include "bluetooth/bluetooth.h"
#include "bluetooth/hci.h"
#include "bluetooth/hci_lib.h"
#include "bluetooth/sdp.h"
#include "bluetooth/sdp_lib.h"

#include "btconstants.h"
#include "btfunctions.h"


bdaddr_t * my_str2ba(char *str) {
	bdaddr_t *addr = NULL;
	addr = malloc(sizeof(bdaddr_t));
	memset(addr,'\0',sizeof(bdaddr_t));
	sscanf(str,"%02X:%02X:%02X:%02X:%02X:%02X",(unsigned int*)&(addr->b[0]),(unsigned int*)&(addr->b[1]),(unsigned int*)&(addr->b[2]),(unsigned int*)&(addr->b[3]),(unsigned int*)&(addr->b[4]),(unsigned int*)&(addr->b[5]));
//	printf("Addr: %02X:%02X:%02X:%02X:%02X:%02X\n",addr->b[0],addr->b[1],addr->b[2],addr->b[3],addr->b[4],addr->b[5]);
	return addr;
}

char * my_ba2str(bdaddr_t *addr) {
	char *str = NULL;
	str = malloc(19);
	memset(str,'\0',19);
	snprintf(str,18,"%02X:%02X:%02X:%02X:%02X:%02X",addr->b[0],addr->b[1],addr->b[2],addr->b[3],addr->b[4],addr->b[5]);
	return str;
}


int my_bacmp(bdaddr_t *addr1, bdaddr_t *addr2) {
	int i = -1;
	i = memcmp(addr1,addr2,sizeof(bdaddr_t));
	return i;
}

void load_blocked_devices(founddevices_t **f) {
	char *query = "SELECT mac FROM content_delivered GROUP BY mac";
	struct sqlite *db;
	device_t *device;
	struct sqlite_vm *pVm;
	const char *sqltail;
	int pN,i=0;
	const char **pazValue;
	const char **pazColName;
	char *errmsg=NULL;
	bdaddr_t ba;
	bdaddr_t *addr_num=NULL;
	
	db = sqlite_open("content_delivered.db", 0, NULL);
	
	if (sqlite_compile(db,query,&sqltail,&pVm,&errmsg) == SQLITE_OK) {
		while (sqlite_step(pVm,&pN,&pazValue,&pazColName) == SQLITE_ROW) {
			addr_num = my_str2ba((char*)(pazValue[0]));
			memcpy(&ba,addr_num,sizeof(bdaddr_t));
			free(addr_num);

			if (is_already_in_list(f, &ba) == FALSE) {
				device = get_device(NULL, NULL, 0, &ba);
				add_remote_device(f, device, STATUS_FINISHED);
				i++;
			}
		}
		sqltail = NULL;
	} else {
		printf("sqlite_compile failed!\nError: %s\n",errmsg);
		free(errmsg);
		errmsg = NULL;
	}
	sqlite_finalize(pVm,NULL);
	pVm = NULL;
	sqlite_close(db);
	printf("Added %u already pushed devices.\n",i);
}


mydevices_t * add_local_device(mydevices_t *mydevs, int dev_id, char purpose) {
	mydevices_t *newdev;
	char tmp[128];
	int sock;
	
	memset(tmp,'\0',128);
	snprintf(tmp,sizeof(tmp),"%s-%i",DEVICE_NAME_PREFIX,dev_id);
	newdev = malloc(sizeof(mydevices_t));
	memset(newdev,'\0',sizeof(mydevices_t));
	
	if(hci_devinfo(dev_id, &(newdev->di)) < 0) {
		/*
		 * Something was not okay,
		 * free the memory and return the old ptr
		 */
		free(newdev);
		return mydevs;
	}
	
	newdev->dev_id 	= dev_id;
	newdev->purpose = purpose;
	newdev->status	= STATUS_READY;
	newdev->tstamp	= time(NULL);
	newdev->thread_id = -1;
	newdev->next 	= mydevs;
	newdev->name	= malloc(strlen(tmp)+1);
	memset(newdev->name,'\0',strlen(tmp)+1);
	strncpy(newdev->name,tmp,strlen(tmp));
	
	sock = hci_open_dev( dev_id );
	if (hci_write_local_name(sock, newdev->name, 1000)<0) {
		printf("Could not set local name to: %s\n",newdev->name);
	} else {
		printf("Set name for device %i to: %s (",newdev->dev_id,newdev->name);
		if (newdev->purpose == PURPOSE_SCAN) printf("SCAN)\n");
		if (newdev->purpose == PURPOSE_PUSH) printf("PUSH)\n");
	}
	close(sock);
	
	return newdev;
}

void add_remote_device(founddevices_t **f, device_t *device, char status) {
	founddevices_t *newdev;
	newdev = malloc(sizeof(founddevices_t));
	memset(newdev,'\0',sizeof(founddevices_t));

//	printf("Adding remote device: %s\n",batostr(&(device->addr)));
	
	newdev->status 	= status;
	newdev->tstamp 	= time(NULL);
	newdev->d		= device;
	newdev->next	= *f;
	*f 				= newdev;
}


void add_locked_device(lockeddevices_t **l, bdaddr_t *addr) {
	lockeddevices_t *newdev;
	struct sembuf sb;

//	printf("Adding locked device: %s\n",batostr(addr));

/*	printf("add_locked_device start\n");
	dump_locked_devices(l);
*/
	sb.sem_num 	= SEM_LOCKSLIST;
	sb.sem_op	= -1;
	sb.sem_flg	= SEM_UNDO;
	semop(semid, &sb, 1);

//	printf("Adding locked device: %s (%p)\n",batostr(addr),addr);
//	if (!addr) printf("FATAL: add_locked_device: addr is NULL\n");
	newdev = malloc(sizeof(lockeddevices_t));
	memset(newdev,'\0',sizeof(lockeddevices_t));
	
	newdev->tstamp 	= time(NULL);
	memcpy(&(newdev->addr),addr,sizeof(bdaddr_t));
	newdev->next = *l;
	*l = newdev;
	
	sb.sem_op = 1;
	semop(semid, &sb, 1);

/*	printf("add_locked_device end\n");
	dump_locked_devices(l);
*/
}


void remove_locked_device(lockeddevices_t **l, bdaddr_t *addr) {
	lockeddevices_t *tmp = *l;
	lockeddevices_t *last = NULL, *next = NULL;
	struct sembuf sb;

	sb.sem_num 	= SEM_LOCKSLIST;
	sb.sem_op	= -1;
	sb.sem_flg	= SEM_UNDO;
	semop(semid, &sb, 1);

//	printf("Removing locked device: %s\n",batostr(addr));
/*	printf("remove_locked_device start\n");
	dump_locked_devices(l);
*/
	while (tmp) {
		next = tmp->next;
		if (!addr) {
			printf("FATAL: remove_locked_device: addr is NULL\n");
		}
		if (!tmp) {
			printf("FATAL: remove_locked_device: tmp is NULL\n");
		}
		if (&(tmp->addr) == NULL) {
			printf("FATAL: remove_locked_device: tmp->addr is NULL\n");
		}
//
//		printf("remove_locked_device(%p,%p)\n",&(tmp->addr),addr);
		if (memcmp((void*)&(tmp->addr),(void*)addr,sizeof(bdaddr_t)) == 0) {
			if (last == NULL) { // item for deletion is first element in list
				*l = tmp->next;
			} else {
				last->next = tmp->next;
			}
			memset(tmp,'\0',sizeof(lockeddevices_t));
			free(tmp);
			tmp=NULL;
		} else {
			last = tmp;
		}
		tmp = next;
	}
	sb.sem_op = 1;
	semop(semid, &sb, 1);
/*	printf("remove_locked_device end\n");
	dump_locked_devices(l);
*/
}





int is_locked(lockeddevices_t **l, bdaddr_t *addr) {
	lockeddevices_t *tmp = *l;
	struct sembuf sb;
	
	sb.sem_num 	= SEM_LOCKSLIST;
	sb.sem_op	= -1;
	sb.sem_flg	= SEM_UNDO;
	semop(semid, &sb, 1);
	
//	printf("Checking locked device: %s\n",batostr(addr));
	for (;tmp;tmp=tmp->next) {
		if (memcmp((void*)&(tmp->addr),(void*)addr,sizeof(bdaddr_t)) == 0) {
//			printf("*** LOCKED\n");
			sb.sem_op = 1;
			semop(semid, &sb, 1);
			return TRUE;
		}
	}
//	printf("*** AVAIL\n");

	sb.sem_op = 1;
	semop(semid, &sb, 1);

	return FALSE;
}


void dump_remote_devices(founddevices_t **f) {
	founddevices_t *tmp = *f;
//	device_caps_t 			*dc;
	char *addr_str = NULL;
	printf("***************************************************\n");
	printf("** DUMPING REMOTE DEVICES                        **\n");
	printf("***************************************************\n");
	if (tmp == NULL) {
		printf("Empty list.\n");
	}
	for (;tmp;tmp=tmp->next) {
		printf("*** STATUS(%i) *** TIME(%i) *********************\n",tmp->status,(int)tmp->tstamp);
		if (tmp->d) {
			printf("Name:  %s\n",tmp->d->name);
			addr_str = my_ba2str(&(tmp->d->addr));
			printf("BDAd:  %s\n",addr_str);
			free(addr_str);
			printf("ScrX:  %i\n",tmp->d->screen_x);
			printf("ScrY:  %i\n",tmp->d->screen_y);
			printf("BipP:  %i\n",tmp->d->bpp);
/*			printf("Caps: ");
			dc = tmp->d->caps;
			for ( ; dc; dc=dc->next) {
				printf("%s,",dc->mime);
			}
*/
			printf("\n");
		}
 
	}
}


founddevices_t * get_remote_device(founddevices_t **f, char status, time_t min_gap) {
	founddevices_t *tmp = *f;
	time_t 			t	= time(NULL);
//	char	*addr_str	= NULL;
	for (;tmp;tmp=tmp->next) {
		if (((t - tmp->tstamp)>min_gap)) { //  && tmp->retries>0) || tmp->retries==0
/*			addr_str = my_ba2str(&(tmp->d->addr));
			printf ("Get remotedev: %s %i\n",addr_str,tmp->status);
			free(addr_str);
*/
			if (tmp->status == status && tmp->d->is_phone == 1) {
				if (tmp->retries > DEVICE_RETRY_TIMOUT) {
					tmp->status = STATUS_TIMEOUT;
				} else {
					return tmp;
				}	
			}
		}
	}
	return NULL;
}


void cleanup_expired_remote_devices(founddevices_t **f, lockeddevices_t **l, time_t min_gap) {
	founddevices_t *tmp = *f,*last=NULL,*next=NULL;
	time_t 			t	= time(NULL);
	struct sembuf 	sb;
	bdaddr_t		a;
	
//	printf("START: Cleaning up old devices\n");
	while (tmp) {
		memset(&a,'\0',sizeof(bdaddr_t));
		memcpy(&a,&(tmp->d->addr),sizeof(bdaddr_t));
		next = tmp->next;
		if ((t - tmp->tstamp)>min_gap ) {
			sb.sem_num 	= SEM_LOCKS;
			sb.sem_op	= -1;
			sb.sem_flg	= SEM_UNDO;
			semop(semid, &sb, 1);
			if (is_locked(l,&a) == FALSE) {
				add_locked_device(l,&a);
				sb.sem_op = 1;
				semop(semid, &sb, 1);
//				printf("Deleting device: %s (P:%p)\n",tmp->d->name,&a);
				if (last == NULL) { // tmp is first item in list
					*f = tmp->next;
				} else {
					last->next = tmp->next;
				}
				if (tmp->d) {
//					printf("Freeing device\n");
					free_device(tmp->d);
					tmp->d=NULL;
				}
				memset(tmp,'\0',sizeof(founddevices_t));
				free(tmp);
				tmp = NULL;
				remove_locked_device(l,&a);
			} else {
				last = tmp;
				sb.sem_op = 1;
				semop(semid, &sb, 1);
			}
		} else {
			last = tmp;
		}
		tmp = next;
	}
//	printf("END: Cleaning up old devices\n");
}



void dump_local_devices(mydevices_t *mydevs) {
	mydevices_t *tmp = mydevs;
	char *addr_str = NULL;
	printf("***************************************************\n");
	printf("** DUMPING LOCAL DEVICES                         **\n");
	printf("***************************************************\n");
	if (tmp == NULL) {
		printf("Empty list.\n");
	}
	for (;tmp;tmp=tmp->next) {
		addr_str = my_ba2str(&(tmp->di.bdaddr));
		printf("Device:\n\tName:%s\n\tAddr:%s\n\tStat:%i\n\tTime:%i\n\tPurp:%s\n\n",
				tmp->name,
				addr_str,
				tmp->status, 
				(int)tmp->tstamp,
				(tmp->purpose==PURPOSE_SCAN ? "SCAN" : "PUSH")
		);
		free(addr_str);
	}
}



void dump_locked_devices(lockeddevices_t **l) {
	lockeddevices_t *tmp = *l;
	char *addr_str = NULL;
	printf("***************************************************\n");
	printf("** DUMPING LOCKED DEVICES                         **\n");
	printf("***************************************************\n");
	if (tmp == NULL) {
		printf("Empty list.\n");
	}
	for (;tmp;tmp=tmp->next) {
		addr_str = my_ba2str(&(tmp->addr));		
		printf("%s (->%p) --- ",addr_str,tmp->next);
		free(addr_str);
	}
	printf("\n\n");
}



mydevices_t * get_free_local_device(mydevices_t *mydevs, char purpose) {
	mydevices_t *tmp = mydevs;
	for (;tmp;tmp=tmp->next) {
		if (tmp->purpose == purpose) {
			if (tmp->status == STATUS_READY) {
//				printf("Free device: %s for %s\n",batostr(&(tmp->di.bdaddr)),(purpose==PURPOSE_SCAN ? "SCAN" : "PUSH"));
				return tmp;
			} else {
				// This device is still locked, it must be an accident
				if ((time(NULL)-tmp->tstamp) > DEVICE_LOCK_TIMEOUT) {
					printf("(hci%i): !!! Uuups. One thread got a timeout! I'm cancelling. (purpose: %s)\n",tmp->dev_id,(purpose==PURPOSE_SCAN ? "SCAN" : "PUSH"));
					pthread_cancel(tmp->thread_id);
					tmp->tstamp = time(NULL);
					tmp->status = STATUS_READY;
					return tmp;
				}
			}
		}
	}
	return NULL;
}


int is_already_in_list(founddevices_t **f, bdaddr_t *addr) {
	founddevices_t *tmp = *f;
//	printf("Checking already_in_list device: %s\n",batostr(addr));
	for (;tmp;tmp=tmp->next) {
		if (tmp->d) {
			if (memcmp((void*)&(tmp->d->addr),(void*)addr,sizeof(bdaddr_t)) == 0) {
				// Renew the timestamp to avoid that the device 
				// will be removed from the list too early
				tmp->tstamp = time(NULL);
				return TRUE;
			}
		}
	}
	return FALSE;
}





int is_local_device (mydevices_t *mydevs, bdaddr_t *addr) {
	mydevices_t *tmp = mydevs;
	
	if (addr == NULL) {
		printf("FATAL: addr was NULL\n");
		return FALSE;
	}
	for (;tmp;tmp=tmp->next) {
		if (my_bacmp(addr,&(tmp->di.bdaddr)) == 0) {
			return TRUE;
		}
	}
	return FALSE;
}

void free_locked_devices(lockeddevices_t **l0, lockeddevices_t *l1) {
	if (l1) {
		if (l1->next)
			free_locked_devices(l0,l1->next);
		free(l1);
		l1=NULL;
	}
}

void free_local_devices(mydevices_t *d) {
	if (d) {
		if (d->next) 
			free_local_devices(d->next);
		if (d->name) {
			free(d->name);
			d->name=NULL;
		}
		memset(d,'\0',sizeof(mydevices_t));
		free(d);
		d=NULL;
	}
}

int get_content_for_device(device_t *d, char *contentid, char *fileprefix, char *usefile, char *usemime, char *useext) {
	char query[4096];
	struct sqlite *db;
	struct sqlite_vm *pVm;
	const char *sqltail;
	int pN;
	const char **pazValue;
	const char **pazColName;
	char *errmsg=NULL;
	int push_default = 1;
	device_caps_t *dc = d->caps;
		
	db = sqlite_open("content.db", 0, NULL);
	
	// Device is not know, query for default content
	if (d->ua_id == 0) {
		push_default = 1;
	} else {
		push_default = 1;
		while (dc) {
			snprintf(query, 4096, "SELECT mime,\"group\",ext FROM content_groups WHERE contentID='%s' AND \"group\"='%s'",contentid,dc->group );			
			if (sqlite_compile(db,query,&sqltail,&pVm,&errmsg) == SQLITE_OK) {
				if (sqlite_step(pVm,&pN,&pazValue,&pazColName) == SQLITE_ROW) {
					memset(usefile,'\0',strlen(fileprefix)+strlen(pazValue[1])+2);
					strcpy(usefile,fileprefix);
					strcat(usefile,"_");
					strcat(usefile,pazValue[1]);
					memset(usemime,'\0',strlen(pazValue[0])+1);
					strcpy(usemime,pazValue[0]);
				memset(useext,'\0',strlen(pazValue[2])+1);
				strcpy(useext,pazValue[2]);

					push_default = 0;
					break;
				}
			} else {
				printf("sqlite_compile failed!\nError: %s\n",errmsg);
				free(errmsg);
				errmsg = NULL;
			}
			sqlite_finalize(pVm,NULL);
			pVm = NULL;
			dc = dc->next;
		}
	}

	if (push_default == 1) {
		snprintf(query, 4096, "SELECT mime,\"group\",ext FROM content_groups WHERE contentID='%s' AND isdef='1'",contentid );
		if (sqlite_compile(db,query,&sqltail,&pVm,&errmsg) == SQLITE_OK) {
			if (sqlite_step(pVm,&pN,&pazValue,&pazColName) == SQLITE_ROW) {
				memset(usefile,'\0',strlen(fileprefix)+strlen(pazValue[1])+2);
				strcpy(usefile,fileprefix);
				strcat(usefile,"_");
				strcat(usefile,pazValue[1]);
				memset(usemime,'\0',strlen(pazValue[0])+1);
				strcpy(usemime,pazValue[0]);
				memset(useext,'\0',strlen(pazValue[2])+1);
				strcpy(useext,pazValue[2]);
			}
		} else {
			printf("sqlite_compile failed!\nError: %s\n",errmsg);
			free(errmsg);
			errmsg = NULL;
		}
		sqlite_finalize(pVm,NULL);
		pVm = NULL;
	}
	sqlite_close(db);

	
	if (usemime != NULL && usefile != NULL) {
//		printf("Returning: %s (%s)\n",usefile,usemime);
		return 0;
	} else {
		return 1;
	}
}

/*
int supports_content_type(device_t *d, char *content_type) {
	device_caps_t *dc= d->caps;
	if (strlen(content_type)<3) return FALSE;
	for (;dc;dc = dc->next) {
		if(strcmp(dc->mime,content_type)==0) {
			return TRUE;
		}
	}
	return FALSE;
}
*/
