#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include "bluetooth/bluetooth.h"
#include "bluetooth/hci.h"
#include "bluetooth/hci_lib.h"
#include "bluetooth/sdp.h"
#include "bluetooth/sdp_lib.h"
#include "sqlite.h"

#include "btconstants.h"
#include "btfunctions.h"


/*
 * Checks the sqlite DB for a devices matching the chunks
 * Returns the ID of the device
 * 
 */
device_t * get_device(bp_hci_chunks_t *h, bp_sdp_chunks_t *s, int channel, bdaddr_t *addr) {
	/*
	 * 1. Must match hci_mac, hci_lmp, hci_features and opush_channel
	 * 2. Select matching records
	 * 3. Check scanned sdp_chunks against record sdp_chunks
	 * 4. Must match at least 75% of record sdp_chunks.
	 */
	char *base_query = "SELECT U.ID,U.Vendor,U.Model,U.ScreenSizeX,U.ScreenSizeY,U.BitsPerPixel,BU.opush_channel,BU.sdp_chunk0,BU.sdp_chunk1,BU.sdp_chunk2,BU.sdp_chunk3,BU.sdp_chunk4,BU.sdp_chunk5,BU.sdp_chunk6,BU.sdp_chunk7,BU.sdp_chunk8,BU.sdp_chunk9,BU.sdp_chunk10,BU.sdp_chunk11,BU.sdp_chunk12,BU.sdp_chunk13,BU.sdp_chunk14,BU.sdp_chunk15,BU.sdp_chunk16,BU.sdp_chunk17,BU.sdp_chunk18,BU.sdp_chunk19,BU.opush_chunk,BU.hci_mac FROM btua AS BU,ua AS U WHERE BU.uaID=U.ID ";
	char tmp[255];
	char query[4096];
	struct sqlite *db;
	struct sqlite_vm *pVm;
	const char *sqltail;
	int pN,num_rows=-1,i=0,j=0;
	const char **pazValue;
	const char **pazColName;
	char *errmsg=NULL;
	device_t *d;
	device_caps_t *dc;
	char chunks_given = 0;
	char chunks_returned = 0;
	unsigned int lastmatch=0, newmatch=0, matches=0;

	d = malloc(sizeof(device_t));
	memset(query,'\0',1024);
	memset(tmp,'\0',255);
	memset(d,'\0',sizeof(device_t));
	memcpy(&(d->addr),addr,sizeof(bdaddr_t));
	d->channel = channel;
	
	if (h == NULL && s == NULL) {
		snprintf(d->ua_id,32,"0%c",0);
		return d;
	}
	
	for (i=0;i<20;i++) {
		if (s->chunk[i]) {
			chunks_given++;
		}
	}

 	snprintf(tmp,255," AND BU.hci_lmp='%s' AND BU.hci_features='%s' AND BU.opush_channel='%i'",h->lmp,h->features,channel);
	strcat(query,base_query);
	strcat(query,tmp);
	
	db = sqlite_open("devices.db", 0, NULL);
	
	if (sqlite_compile(db,query,&sqltail,&pVm,&errmsg) == SQLITE_OK) {
		num_rows = 0;
		while (sqlite_step(pVm,&pN,&pazValue,&pazColName) == SQLITE_ROW) {
			chunks_returned = 0;
			for(i=7;i<27;i++) {
				if (strlen(pazValue[i])!=0) 
					chunks_returned++;
			}
			
//			printf("Comparing: given:%i returned:%i\n",chunks_given,chunks_returned);
			// Compare, 80% must match
			matches=0;
			for (j=0;j<chunks_returned;j++) {
				for (i=0;i<chunks_given;i++) {
					if (strlen(s->chunk[i]) == strlen(pazValue[7+j])) {
						if(memcmp(s->chunk[i],pazValue[7+j],strlen(s->chunk[i])) == 0) {
							matches++;
							break;
						}
					}
				}
			}
			
//			printf("Got %i matches.\n",matches);
			
			if (chunks_returned > chunks_given) {
				newmatch = (unsigned int)(((float)matches / (float)chunks_returned) * 100);
			} else {
				newmatch = (unsigned int)(((float)matches / (float)chunks_given) * 100);
			}
			
//			printf("Comparing: %s == %s\n",s->opush,pazValue[27]);
			if (chunks_returned == 0 && chunks_given==0) {
				if(memcmp(s->opush,pazValue[27],strlen(s->opush)) == 0) {
					newmatch = 90;
					printf("Special case: No SDP chunks available for this device, but Opush SDP chunk okay!\n");
				}
			}
			
			// Comparing if hci_mac is matching 
			// 
			if (memcmp(h->mac,pazValue[28],strlen(h->mac)) == 0) {
				newmatch += 20;
			}
			
			if (newmatch > lastmatch && newmatch >= 70) {
				lastmatch = newmatch;
				if (d->name) free(d->name);
				d->name = malloc(strlen(pazValue[1])+strlen(pazValue[2])+3);
				memset(d->name,'\0',strlen(pazValue[1])+strlen(pazValue[2])+3);
				strcat(d->name,pazValue[1]);
				strcat(d->name," \0");
				strcat(d->name,pazValue[2]);
				d->screen_x = atoi(pazValue[3]);
				d->screen_y = atoi(pazValue[4]);
				d->bpp		= atoi(pazValue[5]);
				d->channel	= atoi(pazValue[6]);
				d->ua_match = lastmatch;
				memset(d->ua_id,'\0',32);
				strncpy(d->ua_id,pazValue[0],32);
				num_rows++;
			}
		}
		sqlite_finalize(pVm,NULL);
		pVm = NULL;
		sqltail = NULL;
		
//		printf("Match score: %i (Device: %s ID: %s)\n",d->ua_match,d->name,d->ua_id);
		
	} else {
		printf("sqlite_compile failed!\nError: %s\n",errmsg);
		free(errmsg);
		free(d);
		errmsg = NULL;
		sqlite_finalize(pVm,NULL);
		pVm = NULL;
		sqlite_close(db);
		return NULL;
	}
	
	
	if (num_rows > 0) {
//		printf ("Device found: %s!\n",d->name);	
		// Querying mime types:
		memset(query,'\0',1024);
		memset(tmp,'\0',255);
		snprintf(query,4096,"SELECT \"group\" FROM ua_groups WHERE uaID='%s' ORDER BY prio DESC",d->ua_id);
//		printf("Query: %s\n",query);
		if (sqlite_compile(db,query,&sqltail,&pVm,&errmsg) == SQLITE_OK) {
			num_rows = 0;
			while (sqlite_step(pVm,&pN,&pazValue,&pazColName) == SQLITE_ROW) {
				dc = malloc(sizeof(device_caps_t));
				memset(dc,'\0',sizeof(device_caps_t));
				dc->group = malloc(strlen(pazValue[0])+1);
				memset(dc->group,'\0',strlen(pazValue[0])+1);
				strcat(dc->group,pazValue[0]);
				
				dc->next = d->caps;
				d->caps = dc;
			}
			sqltail = NULL;
		}
		sqlite_finalize(pVm,NULL);
		pVm = NULL;
		sqlite_close(db);
		return d;
	} else {
		if (num_rows == 0) {
			sqlite_close(db);
			// device was not recognized, insert into unknowndev.db
			db = sqlite_open("unknowndev.db", 0, NULL);
			/*
			 * ID bigint(20) NOT NULL,
			 * defaultname varchar(255) default '',
			 * hci_mac varchar(10) default '',
			 * hci_lmp varchar(20) default '',
			 * hci_features varchar(20) default '',
			 * sdp_chunk0 text,
			 * opush_chunk text NOT NULL,
			 * opush_channel tinyint(4) default '0',
			 * added datetime default '0000-00-00 00:00:00',
			 * uaID bigint(20)
			 */
			snprintf(query, 4096, "INSERT INTO btua VALUES ((SELECT max(ID) FROM btua)+1, '%s', '%s', '%s', '%s'",h->name,h->mac,h->lmp,h->features );
			for (i=0;i<20;i++) {
				if (s->chunk[i]) {
					strcat(query,",'");
					strcat(query,s->chunk[i]);
					strcat(query,"'");
				} else {
					strcat(query,",NULL");
				}
			}
			snprintf(tmp,255,",'%s',%i",s->opush,s->opush_channel);
			strcat(query,tmp);
			strcat(query,",'0000-00-00 00:00:00',0)");
			sqlite_exec(db, query, NULL, NULL, &errmsg);
			if (errmsg) {
				printf("Error: %s\n", errmsg);
				free (errmsg);
			}
			sqlite_close(db);
			
			d->name = malloc(strlen(h->name)+3);
			memset(d->name,'\0',strlen(h->name)+3);
			strcat(d->name,h->name);
			strcat(d->name," \0");
			d->channel	= s->opush_channel;
			memset(d->ua_id,'\0',32);
			snprintf(d->ua_id,32,"%i",0);
			
			printf ("!!!New device (%s) !!! Inserted into unknowndev.db!\n",h->name);
			return d;
		} else {
			sqlite_close(db);
			free(d);
			return NULL;
		}
	}
}


void free_device(device_t *d) {
	if (d->name) { free(d->name); d->name=NULL; }
	if (d->caps) {
		free_device_caps(d->caps);
	}
	memset(d,'\0',sizeof(device_t));
	free(d);
	d=NULL;
}


void free_device_caps(device_caps_t *dc) {
	if (dc->next) 
		free_device_caps(dc->next);
	if (dc->group) {
		free(dc->group);
	}
	memset(dc,'\0',sizeof(device_caps_t));
	free(dc);
	dc=NULL;
}


bp_sdp_t *add_to_bp_list(bp_sdp_t *_bp_list, unsigned int attrId, unsigned char datatype, void *data, unsigned int datasize) {
	bp_sdp_t *new_bp_item;
	new_bp_item = (bp_sdp_t *)malloc(sizeof(bp_sdp_t));
	memset(new_bp_item, '\0', sizeof(bp_sdp_t));

	new_bp_item->datatype	= datatype;
	new_bp_item->attrId		= attrId;
	if (datasize>255) {
		new_bp_item->data = (void*)malloc(255);
		memset(new_bp_item->data, '\0', 255);
		new_bp_item->datalength = 255;
		memcpy(new_bp_item->data,data,255);
	} else if (datasize==0) {
	} else {
		new_bp_item->data = (void*)malloc(datasize);
		memset(new_bp_item->data, '\0', datasize);
		new_bp_item->datalength = datasize;
		memcpy(new_bp_item->data,data,datasize);
	}
	new_bp_item->next = _bp_list;
	
	return new_bp_item;
}

void free_bp_list(bp_sdp_t *_bp_list) {
	if ( _bp_list ) {
		free_bp_list( _bp_list->next );
		if ( _bp_list->data ) {
			free( _bp_list->data );
			_bp_list->data = NULL;
		}
		free( _bp_list );
		_bp_list = NULL;
	}
}


void print_sdp_id(bp_sdp_t *_bp_list) {
	int chunks=0;
	char *sdpstr;
	char tmp[4096];
	char str_attrid[1024]; 		// Values are HEX unsigned short
	char str_rechandle[1024]; 	// Values are HEX unsigned int
	char str_serviceclass[1024];// Values are HEX unsigned short
	char str_protocol[1024];	// Values are HEX unsigned short
	char str_language[1024];
	int last_attrid=-1;
	unsigned char last_rfcomm=0;
	sdpstr = (char*)malloc(255*11);

	printf("\nSDP print:\n");
	memset(sdpstr,'\0',255*11);
	memset(str_attrid,'\0',1024);
	memset(str_rechandle,'\0',1024);
	memset(str_serviceclass,'\0',1024);
	memset(str_protocol,'\0',1024);
	memset(str_language,'\0',1024);
	for (; _bp_list; _bp_list = _bp_list->next) {		
		if ( _bp_list->attrId == SDP_SPACER && _bp_list->data==NULL && _bp_list->datatype == SDP_SPACER) {
			printf("Chunk%i: %s-%s-%s-%s-%s\n",chunks,str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
			memset(sdpstr,'\0',255*11);
			memset(str_attrid,'\0',1024);
			memset(str_rechandle,'\0',1024);
			memset(str_serviceclass,'\0',1024);
			memset(str_protocol,'\0',1024);
			memset(str_language,'\0',1024);
			chunks++;
		} else {
			if(_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x8) {
				last_rfcomm = *(unsigned char*)(_bp_list->data);
			}
			
			if (_bp_list->attrId == SDP_ATTR_RECORD_HANDLE && _bp_list->datatype == 0x0A) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%X",*(unsigned int*)(_bp_list->data));
				if (strlen(str_rechandle)>0) 
					strcat(str_rechandle,",");
				strcat(str_rechandle,tmp);
			}
			
			if (_bp_list->attrId == SDP_ATTR_SVCLASS_ID_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_serviceclass)>0) 
					strcat(str_serviceclass,",");
				strcat(str_serviceclass,tmp);
			}

			if (_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_protocol)>0) 
					strcat(str_protocol,",");
				strcat(str_protocol,tmp);
				if (*(unsigned short*)_bp_list->data == 0x0003) {
					if (last_rfcomm > 0) {
						memset(tmp,'\0',4096);
						snprintf(tmp,4096,"(%X)",last_rfcomm);
						strcat(str_protocol,tmp);
					}
				}
			}

			if (_bp_list->attrId == SDP_ATTR_LANG_BASE_ATTR_ID_LIST	&& _bp_list->datatype == 0x9) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_language)>0) 
					strcat(str_language,",");
				strcat(str_language,tmp);
			}

			
			// New attrid in current record set found, add it to attrid_list
			if (last_attrid != _bp_list->attrId) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",_bp_list->attrId);
				if (strlen(str_attrid)>0) 
					strcat(str_attrid,",");
				strcat(str_attrid,tmp);
			}
		}
		last_attrid = _bp_list->attrId;
	}
	
	printf("Chunk%i: %s-%s-%s-%s-%s\n",chunks,str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
	
	free(sdpstr);
	sdpstr = NULL;
}

void print_sdp_id_html(bp_sdp_t *_bp_list, char* prefix) {
	int chunks=0;
	char *sdpstr;
	char tmp[4096];
	char str_attrid[1024]; 		// Values are HEX unsigned short
	char str_rechandle[1024]; 	// Values are HEX unsigned int
	char str_serviceclass[1024];// Values are HEX unsigned short
	char str_protocol[1024];	// Values are HEX unsigned short
	char str_language[1024];
	int last_attrid=-1;
	unsigned char last_rfcomm=0;
	sdpstr = (char*)malloc(255*11);

	printf("\nSDP print:<br>\n");
	memset(sdpstr,'\0',255*11);
	memset(str_attrid,'\0',1024);
	memset(str_rechandle,'\0',1024);
	memset(str_serviceclass,'\0',1024);
	memset(str_protocol,'\0',1024);
	memset(str_language,'\0',1024);
	for (; _bp_list; _bp_list = _bp_list->next) {		
		if ( _bp_list->attrId == SDP_SPACER && _bp_list->data==NULL && _bp_list->datatype == SDP_SPACER) {
			printf("<input type=\"text\" size=\"150\" name=\"%ssdp_chunk%i\" value=\"%s-%s-%s-%s-%s\"><br>\n",prefix,chunks,str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
			memset(sdpstr,'\0',255*11);
			memset(str_attrid,'\0',1024);
			memset(str_rechandle,'\0',1024);
			memset(str_serviceclass,'\0',1024);
			memset(str_protocol,'\0',1024);
			memset(str_language,'\0',1024);
			chunks++;
		} else {
			if(_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x8) {
				last_rfcomm = *(unsigned char*)(_bp_list->data);
			}
			
			if (_bp_list->attrId == SDP_ATTR_RECORD_HANDLE && _bp_list->datatype == 0x0A) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%X",*(unsigned int*)(_bp_list->data));
				if (strlen(str_rechandle)>0) 
					strcat(str_rechandle,",");
				strcat(str_rechandle,tmp);
			}
			
			if (_bp_list->attrId == SDP_ATTR_SVCLASS_ID_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_serviceclass)>0) 
					strcat(str_serviceclass,",");
				strcat(str_serviceclass,tmp);
			}

			if (_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_protocol)>0) 
					strcat(str_protocol,",");
				strcat(str_protocol,tmp);
				if (*(unsigned short*)_bp_list->data == 0x0003) {
					if (last_rfcomm > 0) {
						memset(tmp,'\0',4096);
						snprintf(tmp,4096,"(%X)",last_rfcomm);
						strcat(str_protocol,tmp);
					}
				}
			}

			if (_bp_list->attrId == SDP_ATTR_LANG_BASE_ATTR_ID_LIST	&& _bp_list->datatype == 0x9) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_language)>0) 
					strcat(str_language,",");
				strcat(str_language,tmp);
			}

			
			// New attrid in current record set found, add it to attrid_list
			if (last_attrid != _bp_list->attrId) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",_bp_list->attrId);
				if (strlen(str_attrid)>0) 
					strcat(str_attrid,",");
				strcat(str_attrid,tmp);
			}
		}
		last_attrid = _bp_list->attrId;
	}
	
	printf("<input type=\"text\" size=\"150\" name=\"%ssdp_chunk%i\" value=\"%s-%s-%s-%s-%s\"><br>\n",prefix,chunks,str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
	
	free(sdpstr);
	sdpstr = NULL;
}

void get_sdp_id(bp_sdp_t *_bp_list, bp_sdp_chunks_t *r) {
	int chunks=0;
	char *sdpstr;
	char tmp[4096];
	char str_attrid[1024]; 		// Values are HEX unsigned short
	char str_rechandle[1024]; 	// Values are HEX unsigned int
	char str_serviceclass[1024];// Values are HEX unsigned short
	char str_protocol[1024];	// Values are HEX unsigned short
	char str_language[1024];
	int last_attrid=-1;
	unsigned char last_rfcomm=0;
	sdpstr = (char*)malloc(255*11);

	memset(sdpstr,'\0',255*11);
	memset(str_attrid,'\0',1024);
	memset(str_rechandle,'\0',1024);
	memset(str_serviceclass,'\0',1024);
	memset(str_protocol,'\0',1024);
	memset(str_language,'\0',1024);
	for (; _bp_list; _bp_list = _bp_list->next) {		
		if ( _bp_list->attrId == SDP_SPACER && _bp_list->data==NULL && _bp_list->datatype == SDP_SPACER) {
			memset(sdpstr,'\0',255*11);	
			snprintf(sdpstr,255*11,"%s-%s-%s-%s-%s",str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
			r->chunk[chunks] = (char*)malloc(strlen(sdpstr)+1);
			memset(r->chunk[chunks],'\0',strlen(sdpstr)+1);
			strncpy(r->chunk[chunks],sdpstr,strlen(sdpstr));
			memset(sdpstr,'\0',255*11);
			memset(str_attrid,'\0',1024);
			memset(str_rechandle,'\0',1024);
			memset(str_serviceclass,'\0',1024);
			memset(str_protocol,'\0',1024);
			memset(str_language,'\0',1024);
			chunks++;
		} else {
			if(_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x8) {
				last_rfcomm = *(unsigned char*)(_bp_list->data);
			}
			
			if (_bp_list->attrId == SDP_ATTR_RECORD_HANDLE && _bp_list->datatype == 0x0A) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%X",*(unsigned int*)(_bp_list->data));
				if (strlen(str_rechandle)>0) 
					strcat(str_rechandle,",");
				strcat(str_rechandle,tmp);
			}
			
			if (_bp_list->attrId == SDP_ATTR_SVCLASS_ID_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_serviceclass)>0) 
					strcat(str_serviceclass,",");
				strcat(str_serviceclass,tmp);
			}

			if (_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_protocol)>0) 
					strcat(str_protocol,",");
				strcat(str_protocol,tmp);
				if (*(unsigned short*)_bp_list->data == 0x0003) {
					if (last_rfcomm > 0) {
						memset(tmp,'\0',4096);
						snprintf(tmp,4096,"(%X)",last_rfcomm);
						strcat(str_protocol,tmp);
					}
				}
			}

			if (_bp_list->attrId == SDP_ATTR_LANG_BASE_ATTR_ID_LIST	&& _bp_list->datatype == 0x9) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_language)>0) 
					strcat(str_language,",");
				strcat(str_language,tmp);
			}

			
			// New attrid in current record set found, add it to attrid_list
			if (last_attrid != _bp_list->attrId) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",_bp_list->attrId);
				if (strlen(str_attrid)>0) 
					strcat(str_attrid,",");
				strcat(str_attrid,tmp);
			}
		}
		last_attrid = _bp_list->attrId;
	}
	
	memset(sdpstr,'\0',255*11);	
	snprintf(sdpstr,255*11,"%s-%s-%s-%s-%s",str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
	r->chunk[chunks] = (char*)malloc(strlen(sdpstr)+1);
	memset(r->chunk[chunks],'\0',strlen(sdpstr)+1);
	strncpy(r->chunk[chunks],sdpstr,strlen(sdpstr));	
	
	free(sdpstr);
	sdpstr = NULL;
}


void get_sdp_obex_id(bp_sdp_t *_bp_list, bp_sdp_chunks_t *r, int channel) {
	char *sdpstr;
	char tmp[4096];
	char str_attrid[1024]; 		// Values are HEX unsigned short
	char str_rechandle[1024]; 	// Values are HEX unsigned int
	char str_serviceclass[1024];// Values are HEX unsigned short
	char str_protocol[1024];	// Values are HEX unsigned short
	char str_language[1024];
	int last_attrid=-1;
	unsigned char last_rfcomm=0;
	sdpstr = (char*)malloc(255*11);

	memset(sdpstr,'\0',255*11);
	memset(str_attrid,'\0',1024);
	memset(str_rechandle,'\0',1024);
	memset(str_serviceclass,'\0',1024);
	memset(str_protocol,'\0',1024);
	memset(str_language,'\0',1024);
	for (; _bp_list; _bp_list = _bp_list->next) {		
			if(_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x8) {
				last_rfcomm = *(unsigned char*)(_bp_list->data);
			}
			
			if (_bp_list->attrId == SDP_ATTR_RECORD_HANDLE && _bp_list->datatype == 0x0A) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%X",*(unsigned int*)(_bp_list->data));
				if (strlen(str_rechandle)>0) 
					strcat(str_rechandle,",");
				strcat(str_rechandle,tmp);
			}
			
			if (_bp_list->attrId == SDP_ATTR_SVCLASS_ID_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_serviceclass)>0) 
					strcat(str_serviceclass,",");
				strcat(str_serviceclass,tmp);
			}

			if (_bp_list->attrId == SDP_ATTR_PROTO_DESC_LIST && _bp_list->datatype == 0x19) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_protocol)>0) 
					strcat(str_protocol,",");
				strcat(str_protocol,tmp);
				if (*(unsigned short*)_bp_list->data == 0x0003) {
					if (last_rfcomm > 0) {
						memset(tmp,'\0',4096);
						snprintf(tmp,4096,"(%X)",last_rfcomm);
						strcat(str_protocol,tmp);
					}
				}
			}

			if (_bp_list->attrId == SDP_ATTR_LANG_BASE_ATTR_ID_LIST	&& _bp_list->datatype == 0x9) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",*(unsigned short*)(_bp_list->data));
				if (strlen(str_language)>0) 
					strcat(str_language,",");
				strcat(str_language,tmp);
			}

			
			// New attrid in current record set found, add it to attrid_list
			if (last_attrid != _bp_list->attrId) {
				memset(tmp,'\0',4096);
				snprintf(tmp,4096,"%hX",_bp_list->attrId);
				if (strlen(str_attrid)>0) 
					strcat(str_attrid,",");
				strcat(str_attrid,tmp);
			}
		last_attrid = _bp_list->attrId;
	}
	
	memset(sdpstr,'\0',255*11);	
	snprintf(sdpstr,255*11,"%s-%s-%s-%s-%s",str_attrid,str_rechandle,str_serviceclass,str_protocol,str_language);
	r->opush = (char*)malloc(strlen(sdpstr)+1);
	memset(r->opush,'\0',strlen(sdpstr)+1);
	strncpy(r->opush,sdpstr,strlen(sdpstr));	
	r->opush_channel = channel;
	free(sdpstr);
	sdpstr = NULL;

}


void get_hci_id(bdaddr_t *baddr, struct hci_version *version, unsigned char *features, bp_hci_chunks_t *r) {
	char tmp[128];
	memset(tmp,'\0',128);
	snprintf(tmp,128,"%02X%02X%02X",baddr->b[5],baddr->b[4],baddr->b[3]);
	r->mac = (char*)malloc(sizeof(char)*strlen(tmp)+1);
	memset(r->mac,'\0',strlen(tmp)+1);
	strncpy(r->mac,tmp,strlen(tmp));
	
	memset(tmp,'\0',128);
	snprintf(tmp,128,"%X%X%X%X%X",version->lmp_ver,version->lmp_subver,version->hci_ver,version->hci_rev,version->manufacturer);
	r->lmp = (char*)malloc(sizeof(char)*strlen(tmp)+1);
	memset(r->lmp,'\0',strlen(tmp)+1);
	strncpy(r->lmp,tmp,strlen(tmp));
	
	memset(tmp,'\0',128);
	snprintf(tmp,128,"%02X%02X%02X%02X%02X%02X%02X%02X",features[0],features[1],features[2],features[3],features[4],features[5],features[6],features[7]);	
	r->features = (char*)malloc(sizeof(char)*strlen(tmp)+1);
	memset(r->features,'\0',strlen(tmp)+1);
	strncpy(r->features,tmp,strlen(tmp));
}



void free_hci_chunks(bp_hci_chunks_t *r) {
	if (r->mac) {		free(r->mac); r->mac = NULL; }
	if (r->lmp)	{		free(r->lmp); r->lmp = NULL; }
	if (r->features) {	free(r->features); r->features=NULL; }
	if (r->name) 	{	free(r->name); r->name=NULL; }
	memset(r,'\0',sizeof(bp_hci_chunks_t));
}

void free_sdp_chunks(bp_sdp_chunks_t *r) {
	int i=0;
	for (i=0;i<20;i++) {
		if (r->chunk[i]) {
			free(r->chunk[i]);
			r->chunk[i] = NULL;
		}
	}
	if (r->opush) {
		free(r->opush);
		r->opush=NULL;
	}
	memset(r,'\0',sizeof(bp_sdp_chunks_t));
}



void print_hci_id(
	bdaddr_t *baddr,
	struct hci_version *version, 
	unsigned char *features) 
{
	printf("HCI print:\n");
	printf("%02X%02X%02X-",baddr->b[5],baddr->b[4],baddr->b[3]);
	printf("%X%X%X%X%X-",version->lmp_ver,version->lmp_subver,version->hci_ver,version->hci_rev,version->manufacturer);
	printf("%02X%02X%02X%02X%02X%02X%02X%02X\n",features[0],features[1],features[2],features[3],features[4],features[5],features[6],features[7]);	
}

void print_hci_id_html(
	bdaddr_t *baddr,
	struct hci_version *version, 
	unsigned char *features) 
{
	printf("<br><br><b>HCI print:</b><br>\n");
	printf("MAC: <input type=\"text\" size=\"50\" name=\"hci_mac\" value=\"%02X%02X%02X\"><br>",baddr->b[5],baddr->b[4],baddr->b[3]);
	printf("LMP: <input type=\"text\" size=\"50\" name=\"hci_lmp\" value=\"%X%X%X%X%X\"><br>",version->lmp_ver,version->lmp_subver,version->hci_ver,version->hci_rev,version->manufacturer);
	printf("FEA: <input type=\"text\" size=\"50\" name=\"hci_features\" value=\"%02X%02X%02X%02X%02X%02X%02X%02X\"><br>",features[0],features[1],features[2],features[3],features[4],features[5],features[6],features[7]);
	printf("<br><br>");
}



bp_sdp_t * sdp_get_attribs_from_record(sdp_data_t *pdata, unsigned int attrId, bp_sdp_t *bp_list) {
	
	sdp_data_t *sdpdata = pdata;
	char debug = 0;
	
	switch (sdpdata->dtd) {
		case SDP_DATA_NIL:
		if (debug == 1) printf("SDP_DATA_NIL ");
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_DATA_NIL,NULL,0);
		break;
	case SDP_BOOL:
		if (debug == 1) printf("SDP_BOOL: %X ",sdpdata->val.int8);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_BOOL,&sdpdata->val.int8,sizeof(sdpdata->val.int8));
		break;
	case SDP_UINT8:
		if (debug == 1) printf("SDP_UINT8: %X ",sdpdata->val.uint8);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UINT8,&sdpdata->val.uint8,sizeof(sdpdata->val.uint8));
		break;
	case SDP_UINT16:
		if (debug == 1) printf("SDP_UINT16: %X ",sdpdata->val.uint16);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UINT16,&sdpdata->val.uint16,sizeof(sdpdata->val.uint16));
		break;
	case SDP_UINT32:
		if (debug == 1) printf("SDP_UINT32: %X ",sdpdata->val.uint32);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UINT32,&sdpdata->val.uint32,sizeof(sdpdata->val.uint32));
		break;
	case SDP_UINT64:
		if (debug == 1) printf("SDP_UINT64: %X ",(unsigned int)sdpdata->val.uint64);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UINT64,&sdpdata->val.uint64,sizeof(sdpdata->val.uint64));
		break;
	case SDP_UINT128:
		if (debug == 1) printf("SDP_UINT128: %02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X "
			,sdpdata->val.uint128.data[0]
			,sdpdata->val.uint128.data[1]
			,sdpdata->val.uint128.data[2]
			,sdpdata->val.uint128.data[3]
			,sdpdata->val.uint128.data[4]
			,sdpdata->val.uint128.data[5]
			,sdpdata->val.uint128.data[6]
			,sdpdata->val.uint128.data[7]
			,sdpdata->val.uint128.data[8]
			,sdpdata->val.uint128.data[9]
			,sdpdata->val.uint128.data[10]
			,sdpdata->val.uint128.data[11]
			,sdpdata->val.uint128.data[12]
			,sdpdata->val.uint128.data[13]
			,sdpdata->val.uint128.data[14]
			,sdpdata->val.uint128.data[15]
			);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UINT128,&sdpdata->val.uint128,sizeof(sdpdata->val.uint128));
		break;
	case SDP_INT8:
		if (debug == 1) printf("SDP_INT8: %X ",sdpdata->val.int8);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_INT8,&sdpdata->val.int8,sizeof(sdpdata->val.int8));
		break;
	case SDP_INT16:
		if (debug == 1) printf("SDP_INT16: %X ",sdpdata->val.int16);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_INT16,&sdpdata->val.int16,sizeof(sdpdata->val.int16));
		break;
	case SDP_INT32:
		if (debug == 1) printf("SDP_INT32: %X ",sdpdata->val.int32);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_INT32,&sdpdata->val.int32,sizeof(sdpdata->val.int32));
		break;
	case SDP_INT64:
		if (debug == 1) printf("SDP_INT64: %X ",(unsigned int)sdpdata->val.int64);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_INT64,&sdpdata->val.int64,sizeof(sdpdata->val.int64));
		break;
	case SDP_INT128:
		if (debug == 1) printf("SDP_INT128: %02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X "
			,sdpdata->val.int128.data[0]
			,sdpdata->val.int128.data[1]
			,sdpdata->val.int128.data[2]
			,sdpdata->val.int128.data[3]
			,sdpdata->val.int128.data[4]
			,sdpdata->val.int128.data[5]
			,sdpdata->val.int128.data[6]
			,sdpdata->val.int128.data[7]
			,sdpdata->val.int128.data[8]
			,sdpdata->val.int128.data[9]
			,sdpdata->val.int128.data[10]
			,sdpdata->val.int128.data[11]
			,sdpdata->val.int128.data[12]
			,sdpdata->val.int128.data[13]
			,sdpdata->val.int128.data[14]
			,sdpdata->val.int128.data[15]
			);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_INT128,&sdpdata->val.int128,sizeof(sdpdata->val.int128));
		break;
	
	case SDP_UUID16:
		if (debug == 1) printf("SDP_UUID16: %X ",sdpdata->val.uuid.value.uuid16);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UUID16,&sdpdata->val.uuid.value.uuid16,sizeof(sdpdata->val.uuid.value.uuid16));
		break;
	case SDP_UUID32:
		if (debug == 1) printf("SDP_UUID32: %X ",sdpdata->val.uuid.value.uuid32);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UUID32,&sdpdata->val.uuid.value.uuid32,sizeof(sdpdata->val.uuid.value.uuid32));
		break;
	case SDP_UUID128:
		if (debug == 1) printf("SDP_UUID128: %02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X "
			,sdpdata->val.uuid.value.uuid128.data[0]
			,sdpdata->val.uuid.value.uuid128.data[1]
			,sdpdata->val.uuid.value.uuid128.data[2]
			,sdpdata->val.uuid.value.uuid128.data[3]
			,sdpdata->val.uuid.value.uuid128.data[4]
			,sdpdata->val.uuid.value.uuid128.data[5]
			,sdpdata->val.uuid.value.uuid128.data[6]
			,sdpdata->val.uuid.value.uuid128.data[7]
			,sdpdata->val.uuid.value.uuid128.data[8]
			,sdpdata->val.uuid.value.uuid128.data[9]
			,sdpdata->val.uuid.value.uuid128.data[10]
			,sdpdata->val.uuid.value.uuid128.data[11]
			,sdpdata->val.uuid.value.uuid128.data[12]
			,sdpdata->val.uuid.value.uuid128.data[13]
			,sdpdata->val.uuid.value.uuid128.data[14]
			,sdpdata->val.uuid.value.uuid128.data[15]
			);

		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_UUID128,&sdpdata->val.uuid.value.uuid128,sizeof(sdpdata->val.uuid.value.uuid128));
		break;
	
	case SDP_TEXT_STR8:
		if (debug == 1) printf("SDP_TEXT_STR8: %s ",sdpdata->val.str);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_TEXT_STR8,sdpdata->val.str,strlen(sdpdata->val.str));
		break;
	case SDP_TEXT_STR16:
		if (debug == 1) printf("SDP_TEXT_STR16: %s ",sdpdata->val.str);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_TEXT_STR16,sdpdata->val.str,strlen(sdpdata->val.str));
		break;
	case SDP_TEXT_STR32:
		if (debug == 1) printf("SDP_TEXT_STR32: %s ",sdpdata->val.str);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_TEXT_STR32,sdpdata->val.str,strlen(sdpdata->val.str));
		break;
	case SDP_URL_STR8:
		if (debug == 1) printf("SDP_URL_STR8: %s ",sdpdata->val.str);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_URL_STR8,sdpdata->val.str,strlen(sdpdata->val.str));
		break;
	case SDP_URL_STR16:
		if (debug == 1) printf("SDP_URL_STR16: %s ",sdpdata->val.str);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_URL_STR16,sdpdata->val.str,strlen(sdpdata->val.str));
		break;
	case SDP_URL_STR32:
		if (debug == 1) printf("SDP_URL_STR16: %s ",sdpdata->val.str);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_URL_STR32,sdpdata->val.str,strlen(sdpdata->val.str));
		break;
	
	case SDP_SEQ8:
		if (debug == 1) printf("SDP_SEQ8: %p ",sdpdata->val.dataseq);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_SEQ8,NULL,0);
		bp_list = sdp_get_attribs_from_record(sdpdata->val.dataseq,attrId,bp_list);
		break;
	case SDP_SEQ16:
		if (debug == 1) printf("SDP_SEQ16: %p ",sdpdata->val.dataseq);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_SEQ16,NULL,0);
		bp_list = sdp_get_attribs_from_record(sdpdata->val.dataseq,attrId,bp_list);
		break;
	case SDP_SEQ32:
		if (debug == 1) printf("SDP_SEQ32: %p ",sdpdata->val.dataseq);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_SEQ32,NULL,0);
		bp_list = sdp_get_attribs_from_record(sdpdata->val.dataseq,attrId,bp_list);
		break;
	
	case SDP_ALT8:
		if (debug == 1) printf("SDP_ALT8: %p ",sdpdata->val.dataseq);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_ALT8,NULL,0);
		bp_list = sdp_get_attribs_from_record(sdpdata->val.dataseq,attrId,bp_list);
		break;
	case SDP_ALT16:
		if (debug == 1) printf("SDP_ALT16: %p ",sdpdata->val.dataseq);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_ALT16,NULL,0);
		bp_list = sdp_get_attribs_from_record(sdpdata->val.dataseq,attrId,bp_list);
		break;
	case SDP_ALT32:
		if (debug == 1) printf("SDP_ALT32: %p ",sdpdata->val.dataseq);
		bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,attrId,SDP_ALT32,NULL,0);
		bp_list = sdp_get_attribs_from_record(sdpdata->val.dataseq,attrId,bp_list);
		break;
	}

	
	if (sdpdata->next) {
		bp_list = sdp_get_attribs_from_record(sdpdata->next,attrId,bp_list);
	}

	return bp_list;
}



int sdp_get_opush_channel(sdp_data_t *pdata, unsigned int attrId, int c, char in_section, int *channel) {

	// Path to OPUSH channel:
	// Match: AtrrId == 0x4
	// Get: SEQ->SEQ[x]->Channel

	sdp_data_t *sdpdata = pdata;
	char debug=0;
	if (debug == 1) printf("#");
	switch (sdpdata->dtd) {
	case SDP_UINT8:
	case SDP_UINT16:
	case SDP_UINT32:
	case SDP_UINT64:
	case SDP_UINT128:
		if (debug == 1) printf("SDP_UINTX: %X ",sdpdata->val.uint32);
		c = (int)(sdpdata->val.uint32);
		if (in_section != -1) {
			*channel = c;
		}		
		break;
	case SDP_INT8:
	case SDP_INT16:
	case SDP_INT32:
	case SDP_INT64:
	case SDP_INT128:
		if (debug == 1) printf("SDP_INTX: %X ",sdpdata->val.int32);
		c = (int)sdpdata->val.int32;
		if (in_section != -1) {
			*channel = c;
		}
		break;
	
	case SDP_UUID16:
	case SDP_UUID32:
	case SDP_UUID128:
		if (debug == 1) printf("SDP_UUIDX: %X ",(unsigned int)sdpdata->val.uuid.value.uuid32);
		if (sdpdata->val.uuid.value.uuid32 == 0x0003) {
			in_section=1;
			/*
			 * We already found a INT entry in this section.
			 * Treat this INT as our wanted channel
			 */
			if (c != -1) {
				*channel = c;
			}
		}
		break;
		
	case SDP_SEQ8:
	case SDP_SEQ16:
	case SDP_SEQ32:
	case SDP_ALT8:
	case SDP_ALT16:
	case SDP_ALT32:
		if (debug == 1) printf("SDP_SEQX: %p ",sdpdata->val.dataseq);
		sdp_get_opush_channel(sdpdata->val.dataseq,attrId,-1,0,channel);
		break;
	}
	
	if (sdpdata->next) {
		sdp_get_opush_channel(sdpdata->next,attrId,c,in_section,channel);
	}

	return 0;
}
