#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include "bluetooth/bluetooth.h"
#include "bluetooth/rfcomm.h"
#include "bluetooth/hci.h"

#include "btconstants.h"
#include "btfunctions.h"


int obex_push(char *filename, char *mime, char *ext, mydevices_t *me, founddevices_t *to) {
	int 							sock=-1;
	struct sockaddr_rc 				laddr, raddr;
	unsigned char 					recv_buf[8192];
	int 							recv_bytes = 0, i=0;
	struct obex_connect_response 	ocr;
	int 							maxlen = 0, read_bytes;
	int 							chunk = 0;
	FILE 							*in;
	char							*send_buf;
	int								chunk_size=0,buf_pos=0;
	unsigned long 					filesize=0;
	char							*uc_filename;
	int								uc_filename_length, uc_slash_pos=0;
	bdaddr_t 						*from=&(me->di.bdaddr);
	char 							debug = 0, no_mime=0;
	time_t							start = time(NULL);
	char 							httpheader[2048];
	char							*addr_str;
	char							*newfilename;

	/* CONNECT Final bit set: 	0x80
	 * LENGTH of packet: 		0x0007
	 * VERSION of OBEX: 		0x10
	 * FLAGS none:				0x00
	 * MAX PACKETSIZE:			0x2000
	 */
	unsigned char const data_obex_connect[7] = {
	    128,   0,   7,  16,   0,  32,  0
	};
	
	for (i=strlen(filename);i>=0;i--) {
		if (filename[i] == '/') {
			uc_slash_pos = i+1;
			break;
		}
	}
	
	newfilename = malloc(512);
	memset(newfilename,'\0',sizeof(newfilename));
	strcpy(newfilename,filename);
	strcat(newfilename,".");
	strcat(newfilename,ext);

	uc_filename_length = (strlen(newfilename)+1)*2;
	uc_filename = (char*)malloc(uc_filename_length);
	memset(uc_filename,'\0',uc_filename_length);
	for (i=uc_slash_pos;i<strlen(newfilename);i++) {
		uc_filename[2*i]	= 0;
		uc_filename[2*i+1]	= newfilename[i];
	}

	if (newfilename) {
		free(newfilename);
		newfilename = NULL;	
	}
	
	in = fopen(filename,"r");
	if (in == NULL) {
		printf("File: %s not found.\n",filename);
		free(uc_filename);
		return 1;
	}
	fseek(in, 0, SEEK_END );
    filesize= ftell( in ); 
    fseek(in, 0, SEEK_SET );
    if (debug == 1) 
		printf("File (%s): %li bytes\n",filename,filesize);
	
	snprintf(httpheader,2048,"Content-Type:%s\r\nContent-Length:%lu\r\n",mime,filesize);
	
	memset(&laddr,'\0',sizeof(struct sockaddr_rc));
	memset(&raddr,'\0',sizeof(struct sockaddr_rc));
	
	laddr.rc_family = AF_BLUETOOTH;
	laddr.rc_bdaddr = *from;
	laddr.rc_channel = 0;
	
	raddr.rc_family = AF_BLUETOOTH;
	raddr.rc_bdaddr = to->d->addr;
	raddr.rc_channel = htobs(to->d->channel);
	
	if (debug == 1) {
		addr_str = my_ba2str(&laddr.rc_bdaddr);
		printf("RFCOMM conn from: %s to: %s\n",addr_str,batostr(&raddr.rc_bdaddr));
		free(addr_str);
	}
	
	if((sock = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		to->status = STATUS_FAILED;
		to->tstamp = time(NULL);
		free(uc_filename);
		fclose(in);
		return 1;
	}
	
	if (debug == 1) 
		printf("Opened socket: %i\n",sock);
	
	if(bind(sock, (struct sockaddr *)&laddr, sizeof(laddr)) < 0) {
		to->status = STATUS_FAILED;
		to->tstamp = time(NULL);
		free(uc_filename);
		close(sock);
		fclose(in);
		return 1;
	}
	
	if(connect(sock, (struct sockaddr *)&raddr, sizeof(raddr)) < 0) {
		if (errno == ECONNREFUSED) {
			to->status = STATUS_REJECTED;
			printf("(hci%i): Client rejected request.\n",me->dev_id);
		} else if (errno == ETIMEDOUT || errno == ETIME || errno == EAGAIN) {
			to->retries++;
			to->status = STATUS_READY;
			printf("(hci%i): Client request timed out.\n",me->dev_id);
		} else {
			to->status = STATUS_FAILED;
			printf("(hci%i): Unknown error (%i): %s.\n",me->dev_id,errno,strerror(errno));
		}
		to->tstamp = time(NULL);
		free(uc_filename);
		close(sock);
		fclose(in);
		return 1;
	}
	
	if (debug == 1) 
		printf("Connected.\n");
	if (send(sock,data_obex_connect,7,0) < 0) {
		to->status = STATUS_FAILED;
		to->tstamp = time(NULL);
		free(uc_filename);
		close(sock);
		fclose(in);
		return 1;
	}

	memset(&ocr,'\0',sizeof(ocr));	
	if ((recv_bytes = recv(sock,&ocr,sizeof(ocr),0)) < 0) {
		to->status = STATUS_FAILED;
		to->tstamp = time(NULL);
		free(uc_filename);
		printf("(hci%i): Receiving connect response failed.\n",me->dev_id);
		return 1;
	} else {
		if (debug == 1) 
			printf("Status: %X\nPLen:%i\nOBEX V.:%X\nFlags:%X\nMaxLen:%i\nSizeof:%i\n",ocr.status,ocr.length[0]*256+ocr.length[1],ocr.obex_version,ocr.flags,ocr.maxlen[0]<<8|ocr.maxlen[1],sizeof(ocr));
		maxlen = ocr.maxlen[0]<<8|ocr.maxlen[1];
	}
	
	if (maxlen >= 8192) {
		chunk_size = 8192;
	} else {
		chunk_size = maxlen-uc_filename_length-strlen(mime)-strlen(httpheader)-64;
	}
	if (chunk_size < 0) {
		no_mime = 1;
		chunk_size = maxlen-uc_filename_length-strlen(mime)-64;
		printf("OBEX: Header too large, skipping HTTP header.\n");
	}
	
	send_buf = (char*)malloc(9000);
	if (debug == 1) 
		printf("Chunk size: %i\n",chunk_size);
	
	while (!feof(in)) {
		buf_pos=0;
		if (chunk > 0) {
			chunk_size = maxlen - 8;
		}
		if (debug == 1) 
			printf("Reading chunk %i ...\n",chunk);
		read_bytes = fread(recv_buf,sizeof(char),chunk_size,in);
		if (debug == 1) 
			printf("Read %i bytes from file.\n",read_bytes);
		memset(send_buf,'\0',9000);
		if (read_bytes < chunk_size) { // this packet is final!
			if (debug == 1) 
				printf("Building PUT_FINAL packet.\n");
			send_buf[buf_pos] = (unsigned char)OBEX_PUT_FINAL; buf_pos++;
			send_buf[buf_pos] = (unsigned char)0;  buf_pos++;// LEN1
			send_buf[buf_pos] = (unsigned char)0;  buf_pos++;// LEN2
			if (chunk == 0) {
				send_buf[buf_pos] = OBEX_HI_NAME;  buf_pos++;
				send_buf[buf_pos] = (unsigned char)(((uc_filename_length+3) & 0xFF00) >> 8);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((uc_filename_length+3) & 0x00FF);  buf_pos++;
				memcpy(&send_buf[buf_pos],uc_filename,uc_filename_length);  buf_pos=buf_pos+uc_filename_length;
				send_buf[buf_pos] = OBEX_HI_LENGTH;  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0xFF000000) >> 24);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0x00FF0000) >> 16);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0x0000FF00) >> 8);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0x000000FF));  buf_pos++;
				send_buf[buf_pos] = OBEX_HI_TYPE;  buf_pos++;
				send_buf[buf_pos] = (unsigned char)(((strlen(mime)+4) & 0xFF00) >> 8);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((strlen(mime)+4) & 0x00FF);  buf_pos++;
				memcpy(&send_buf[buf_pos],mime,strlen(mime)+1);  buf_pos=buf_pos+strlen(mime)+1;
				if (no_mime == 0) {
					send_buf[buf_pos] = OBEX_HI_HTTP;  buf_pos++;
					send_buf[buf_pos] = (unsigned char)(((strlen(httpheader)+4) & 0xFF00) >> 8);  buf_pos++;
					send_buf[buf_pos] = (unsigned char)((strlen(httpheader)+4) & 0x00FF);  buf_pos++;
					memcpy(&send_buf[buf_pos],httpheader,strlen(httpheader)+1);  buf_pos=buf_pos+strlen(httpheader)+1;
				}
			}
			send_buf[buf_pos] = OBEX_HI_ENDOFBODY;  buf_pos++;
			send_buf[buf_pos] = (unsigned char)(((read_bytes+3) & 0xFF00) >> 8);  buf_pos++;
			send_buf[buf_pos] = (unsigned char)((read_bytes+3) & 0x00FF);  buf_pos++;
			memcpy(&send_buf[buf_pos],recv_buf,read_bytes);  buf_pos=buf_pos+read_bytes;
			//buf_pos--;
			send_buf[1] = (unsigned char)((buf_pos & 0xFF00) >> 8);
			send_buf[2] = (unsigned char)(buf_pos & 0x00FF);
		} else { // not final more packets will follow
			if (debug == 1) 
				printf("Building PUT_CHUNK packet.\n");
			send_buf[buf_pos] = (char)OBEX_PUT_CHUNK; buf_pos++;
			send_buf[buf_pos] = (char)0;  buf_pos++;// LEN1
			send_buf[buf_pos] = (char)0;  buf_pos++;// LEN2
			if (chunk == 0) { // first packet, send header
				send_buf[buf_pos] = OBEX_HI_NAME;  buf_pos++;
				send_buf[buf_pos] = (unsigned char)(((uc_filename_length+3) & 0xFF00) >> 8);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((uc_filename_length+3) & 0x00FF);  buf_pos++;
				memcpy(&send_buf[buf_pos],uc_filename,uc_filename_length);  buf_pos=buf_pos+uc_filename_length;
				send_buf[buf_pos] = OBEX_HI_LENGTH;  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0xFF000000) >> 24);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0x00FF0000) >> 16);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0x0000FF00) >> 8);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((filesize & 0x000000FF));  buf_pos++;
				send_buf[buf_pos] = OBEX_HI_TYPE;  buf_pos++;
				send_buf[buf_pos] = (unsigned char)(((strlen(mime)+4) & 0xFF00) >> 8);  buf_pos++;
				send_buf[buf_pos] = (unsigned char)((strlen(mime)+4) & 0x00FF);  buf_pos++;
				memcpy(&send_buf[buf_pos],mime,strlen(mime)+1);  buf_pos=buf_pos+strlen(mime)+1;
				if (no_mime == 0) {
					send_buf[buf_pos] = OBEX_HI_HTTP;  buf_pos++;
					send_buf[buf_pos] = (unsigned char)(((strlen(httpheader)+4) & 0xFF00) >> 8);  buf_pos++;
					send_buf[buf_pos] = (unsigned char)((strlen(httpheader)+4) & 0x00FF);  buf_pos++;
					memcpy(&send_buf[buf_pos],httpheader,strlen(httpheader)+1);  buf_pos=buf_pos+strlen(httpheader)+1;
				}
			}
			send_buf[buf_pos] = OBEX_HI_BODYCHUNK;  buf_pos++;
			send_buf[buf_pos] = (unsigned char)(((read_bytes+3) & 0xFF00) >> 8);  buf_pos++;
			send_buf[buf_pos] = (unsigned char)((read_bytes+3) & 0x00FF);  buf_pos++;
			memcpy(&send_buf[buf_pos],recv_buf,read_bytes);  buf_pos=buf_pos+read_bytes;
			//buf_pos--;
			send_buf[1] = (unsigned char)((buf_pos & 0xFF00) >> 8);
			send_buf[2] = (unsigned char)(buf_pos & 0x00FF);
		}
		if (debug == 1) {
			printf("Sending data.\n");
			printf("Dumping ...\n");
			for(i=1;i<=buf_pos;i++) {
				
				printf("%02X ",*(unsigned char*)&send_buf[i-1]);
				if (i % 16 == 0) printf("\n");
			}
		}
		if (send(sock,send_buf,buf_pos,0) < 0) {
			to->status = STATUS_FAILED;
			to->tstamp = time(NULL);
			free(uc_filename);
			free(send_buf);
			close(sock);
			fclose(in);
			printf("(hci%i): Sending data failed.\n",me->dev_id);
			return 1;
		}
		
		if (debug == 1) 
			printf("Receiving data.\n");
		if ((recv_bytes = recv(sock,recv_buf,1024,0)) < 0) {
			to->status = STATUS_FAILED;
			to->tstamp = time(NULL);
			free(uc_filename);
			free(send_buf);
			close(sock);
			fclose(in);
			printf("(hci%i): Receiving data response failed.\n",me->dev_id);
			return 1;
		} else {
			if (debug == 1) {
				printf("Chunk %i: Received %i bytes.\nDumping ...\n",chunk,recv_bytes);
				for (i=0;i<recv_bytes;i++) {
					printf("%02X ",recv_buf[i]);
				}
				printf("\n");
			}
		}
		
//		printf("%i*",me->dev_id);
		chunk++;
	}
	
	to->status = STATUS_FINISHED;
	to->tstamp = time(NULL);
	
	free(uc_filename);
	free(send_buf);
	close(sock);
	fclose(in);
	
	printf("(hci%i): OBEX push OK. %li bytes in %li seconds.\n",me->dev_id,filesize,time(NULL)-start);
	return 0;
}
