#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include "sqlite.h"
#include "bluetooth/bluetooth.h"
#include "bluetooth/hci.h"
#include "bluetooth/hci_lib.h"
#include "bluetooth/sdp.h"
#include "bluetooth/sdp_lib.h"

#include "btconstants.h"
#include "btfunctions.h"



void *thread_scan (void *param) {
	thread_param_t 			*p 	= (thread_param_t *)param;
	founddevices_t 			**f	= (founddevices_t **)p->remote_device_list;
	mydevices_t	   			*m	= (mydevices_t *)p->local_device_list;
	mydevices_t	   			*me	= (mydevices_t *)p->use_this_device;
	lockeddevices_t			**l	= (lockeddevices_t **)p->locked_devices;
    uint32_t 				range = 0x0000ffff;
    struct hci_version 		version;
	struct hci_conn_info_req *cr;
	uint16_t 				handle;
	inquiry_info 			*ii = NULL;
	int 					max_rsp, num_rsp, sock;
	int 					len, flags;
	int 					iIter;
//	char 					addr[20] = { 0 };
	char 					name[248] = { 0 };
	struct search_context 	context;
	sdp_list_t 				*attrid, *search, *seq, *next;
	sdp_session_t 			*sess;
	sdp_record_t 			*rec;
	sdp_list_t				*list;
	sdp_data_t 				*sdpdata;
	uint16_t 				attrId;
	unsigned char			handle_ok;
	uint8_t 				features[8];
	int						channel=-1;
	bp_sdp_chunks_t			*sdpc;
	bp_hci_chunks_t			*hcic;
	device_t 				*device;
	bp_sdp_t	 			*bp_list=NULL;
	char 					failure = FALSE;
	struct sembuf 			sb;
	uint8_t 				cls[3];
	
	len = DEVICE_SCANTIME;
	max_rsp = 10;
	flags = IREQ_CACHE_FLUSH;
	ii = (inquiry_info*)malloc(max_rsp * sizeof(inquiry_info));
//	memset(ii,'\0',max_rsp * sizeof(inquiry_info));

	printf("(hci%i): SCANNING\n",me->dev_id);
	
	//printf("Inquiring...\n");
	num_rsp = hci_inquiry(me->dev_id, len, max_rsp, NULL, &ii, flags);
	if ( num_rsp < 0) {
		printf("(hci%i): hci_inquiry failed (%s)\n",me->dev_id,strerror(errno));
		free(ii);
		me->status = STATUS_READY;
		me->tstamp = time(NULL);
		free(param);
		return NULL;
	}
	
//	printf("Inquiry finished. Found %i devices on dev: %i.\n",num_rsp,me->dev_id);

	for (iIter=0; iIter<num_rsp; iIter++) {
		channel = -1;
		if (is_local_device (m, &(ii+iIter)->bdaddr) == FALSE) {
			sb.sem_num 	= SEM_LOCKS;
			sb.sem_op	= -1;
			sb.sem_flg	= SEM_UNDO;
//			printf("is_already_in_list(%p,%s)\n",f, batostr(&(ii+iIter)->bdaddr)); 
			if (is_already_in_list(f, &(ii+iIter)->bdaddr) == FALSE) {
				semop(semid, &sb, 1);
				if (is_locked(l,&(ii+iIter)->bdaddr) == FALSE) {
//					printf("bluethreads.c(1): Adding locked device: %s\n",batostr(&(ii+iIter)->bdaddr));
					add_locked_device(l,&(ii+iIter)->bdaddr);
					sb.sem_op = 1;
					semop(semid, &sb, 1);
					failure = FALSE;
					sdpc = malloc(sizeof(bp_sdp_chunks_t));
					hcic = malloc(sizeof(bp_hci_chunks_t));
					memset(sdpc,'\0',sizeof(bp_sdp_chunks_t));
					memset(hcic,'\0',sizeof(bp_hci_chunks_t));
					memset(name, 0, sizeof(name));
					sdpc->opush_channel = -1;
			
//					printf("HCI printing\n");
					/**************************************************************
					 * 
					 * HCI Blueprinting for each found device
					 * 
					 **************************************************************/
			
					handle_ok = 0;
					cr = malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
					if (!cr) {
						printf("Allocating memory for struct cr failed!\n");
						failure = TRUE;
					} else {
						memset(cr, '\0', sizeof(*cr) + sizeof(struct hci_conn_info));
						bacpy(&cr->bdaddr, &(ii+iIter)->bdaddr);
						cr->type = ACL_LINK;
						sock = hci_open_dev( me->dev_id );
						if (ioctl(sock, HCIGETCONNINFO, (unsigned long) cr) < 0) {
							if (hci_create_connection(sock, &(ii+iIter)->bdaddr,htobs(me->di.pkt_type & ACL_PTYPE_MASK),0, 0x01, &handle, DEVICE_DEF_TIMEOUT) < 0) {
								remove_locked_device(l,&(ii+iIter)->bdaddr);
								free_sdp_chunks(sdpc);
								free_hci_chunks(hcic);
								free(sdpc);
								free(hcic);
								free(cr);
								free(ii);
								me->status = STATUS_READY;
								me->tstamp = time(NULL);
								free(param);
								return NULL;
								/*
								printf("(hci%i): Added unknown device with hci connect failure.\n",me->dev_id);
								device = get_device(NULL, NULL, 0, &(ii+iIter)->bdaddr);
								add_remote_device(f, device, STATUS_FAILED);
								failure = TRUE;
								*/
								
							} else {
								handle_ok = 1;
							}
						} else {
							handle = htobs(cr->conn_info->handle);
							handle_ok = 1;
						}
						if (handle_ok == 1) {
							if (hci_read_remote_name(sock, &(ii+iIter)->bdaddr, sizeof(name), name,DEVICE_DEF_TIMEOUT) < 0) 
								strcpy(name,"[unknown]");
							memset(&version, '\0', sizeof(struct hci_version));
							hci_read_remote_version(sock, handle, &version, 20000);
							/*
							printf("****************************************************************\n");
							printf("%s: %s\n",addr,name);
							printf("LMP version: 0x%X.0x%X\n",version.lmp_ver,version.lmp_subver);
							printf("HCI version: 0x%X.0x%X\n",version.hci_ver,version.hci_rev);
							printf("Manufacture: 0x%X\n",version.manufacturer);
							*/
							hci_read_remote_features(sock, handle, features, DEVICE_DEF_TIMEOUT);
							//printf("Features   : 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\n",features[0],features[1],features[2],features[3],features[4],features[5],features[6],features[7]);
							
							printf("(hci%i): Found %s\n",me->dev_id,name);
							
							get_hci_id(&(ii+iIter)->bdaddr,&version,features,hcic);
							if (hcic) {
								if (hcic->name) free(hcic->name);
								hcic->name = malloc(strlen(name)+1);
								memset(hcic->name,'\0',strlen(name)+1);
								strncpy(hcic->name,name,strlen(name));
							}
						} 
						free(cr);
						close(sock);
					}
					
					
					
					
//					printf("SDP printing\n");
					/**************************************************************
					 * 
					 * SDP Blueprinting for each found device
					 * 
					 **************************************************************/
					
					if (failure == FALSE) {
						//printf ("SDP inquiry: \n");
						/* Zero out the search context */
						memset(&context, '\0', sizeof(struct search_context));
						sdp_uuid16_create(&context.group, PUBLIC_BROWSE_GROUP);
						sess = sdp_connect(&(me->di.bdaddr), &(ii+iIter)->bdaddr, 0); // SDP_RETRY_IF_BUSY
						if(!sess) {
							printf("(hci%i): No SDP PUBLIC_BROWSE_GROUP available. Skipping.\n",me->dev_id);
							failure = TRUE;
//							perror("Failed to connect to SDP server.");
							/* Ignore this error,
							 * maybe the device does not support SDP_BROWSE
							 */
						} else {
//							printf("Searching SDP services (2)\n");
							attrid = sdp_list_append(0, &range);
							search = sdp_list_append(0, &context.group);
							if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
//								printf("Service Search failed.\n");
							} else {
								//printf("SDP result: \n");
//								printf("Service search finished. Running seq.\n");
								for (; seq; seq = next) {
									rec = (sdp_record_t *) seq->data;
									list = (sdp_list_t *)rec->attrlist;
									for (; list; list = list->next) {
										sdpdata = (sdp_data_t *)list->data;
										attrId = sdpdata->attrId;
										/*
										 * sdp_get_attribs_from_record reads attribs from record
										 * and stores it in an own struct bp_sdp (bpconstants.h).
										 * This struct is a list and it will be finally used
										 * to generate the SDP print
										 */
										bp_list = sdp_get_attribs_from_record(sdpdata,attrId,bp_list);
									}
									if (seq->next) {
										bp_list = (bp_sdp_t *)add_to_bp_list(bp_list,0xFF,SDP_SPACER,NULL,0);
									}
									next = seq->next;
									free(seq);
									sdp_record_free(rec);
								}
//								printf("get_sdp_id(%p,%p)\n",bp_list,sdpc);
								get_sdp_id(bp_list,sdpc);
								//printf("Freeing SDP list\n");
								free_bp_list(bp_list);
								bp_list = NULL;
							}
							sdp_list_free(attrid, 0);
							sdp_list_free(search, 0);							
							sdp_close(sess);
/*							
						}
				
						if (failure == FALSE) {
*/
	//						printf("SDP OBEX printing\n");
							// Searching for OPUSH facility
							memset(&context, '\0', sizeof(struct search_context));
							sdp_uuid16_create(&context.group, OBEX_OBJPUSH_SVCLASS_ID);
							sess = sdp_connect(&(me->di.bdaddr), &(ii+iIter)->bdaddr, 0); // SDP_RETRY_IF_BUSY
							if(!sess) {
	//							perror("Failed to connect to SDP server.");
							} else {
								attrid = sdp_list_append(0, &range);
								search = sdp_list_append(0, &context.group);
								if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
	//								printf("Service Search failed.\n");
								} else {
									for (; seq; seq = next) {
										rec = (sdp_record_t *) seq->data;
										list = (sdp_list_t *)rec->attrlist;
										for (; list; list = list->next) {
											sdpdata = (sdp_data_t *)list->data;
											attrId = sdpdata->attrId;
											/*
											 *  Evaluate only the ProtocolDescriptorList (attrId == 0x04)
											 */
											if (attrId == 0x04) {
												sdp_get_opush_channel(sdpdata,attrId,-1,0,&channel);
											}
											bp_list = sdp_get_attribs_from_record(sdpdata,attrId,bp_list);
										}
										next = seq->next;
										free(seq);
										sdp_record_free(rec);
									}
									/*
									 * print_sdp_id(bp_list);
									 * printf("OPUSH Channel: %i\n\n",channel);
									 */
									get_sdp_obex_id(bp_list,sdpc,channel);
									free_bp_list(bp_list);
									bp_list = NULL;
								}
								sdp_list_free(attrid, 0);
								sdp_list_free(search, 0);
								sdp_close(sess);
							}
						}
						
						if (failure == FALSE) {
							device = get_device(hcic, sdpc, channel, &(ii+iIter)->bdaddr);
						} else {
							device = get_device(NULL, NULL, channel, &(ii+iIter)->bdaddr);
						}	
						
						if (device) {
							memset(&cls,'\0',3);
							memcpy(cls, (ii+iIter)->dev_class, 3);
							if ((cls[1] & 0x1f) == 2) {
								device->is_phone = 1;
								printf("(hci%i): '%s' is a phone.\n",me->dev_id,hcic->name);
							}
							if (atoi(device->ua_id) > 0) {
								printf("(hci%i): Identified '%s' as %s (%u).\n",me->dev_id,hcic->name,device->name,atoi(device->ua_id));
								add_remote_device(f, device, STATUS_READY);
							} else {
								if (PUSH_UNKNOWNDEV == 1 && device->is_phone == 1) {
									add_remote_device(f, device, STATUS_READY);
								} else {
									add_remote_device(f, device, STATUS_UNKNOWNDEV);
								}
							}
						} else {
							printf("device is NULL\n");
						}
						
						free_sdp_chunks(sdpc);
						free_hci_chunks(hcic);
					}
					free(sdpc);
					free(hcic);
					
//					printf("thread_scan: Removing addr at p:%p\n",&(ii+iIter)->bdaddr);
					remove_locked_device(l,&(ii+iIter)->bdaddr);
				} else { // End of is_locked
					sb.sem_op = 1;
					semop(semid, &sb, 1);
				}
			}  // End of is_already_in_list
		} // End of is_local_device 
	} // End of for
	
	free(ii);
	me->status = STATUS_READY;
	me->tstamp = time(NULL);
	
	free(param);
	
	return NULL;
}






void *thread_push (void *param) {
	thread_param_t 			*p 	= (thread_param_t *)param;
	mydevices_t	   			*me	= (mydevices_t *)p->use_this_device;
	lockeddevices_t			**l	= (lockeddevices_t **)p->locked_devices;
//	device_t				*d	= (device_t *)p->use_this_remote_device;
	founddevices_t 			*d	= (founddevices_t *)p->use_this_remote_device;
	char 					query[4096];
	struct sqlite 			*db,*db2;
	char 					*errmsg=NULL;
	struct sqlite_vm 		*pVm;
	const char 				*sqltail;
	int 					pN=0;
	const char 				**pazValue;
	const char 				**pazColName;
	char					*addr_str=NULL;
	char					*usemime=NULL,*usefile=NULL,*useext=NULL;
	int						content_pushed=0;
	
	usemime = malloc(4096);
	usefile = malloc(4096);
	useext = malloc(4096);
	
	if (is_locked(l,&(d->d->addr)) == TRUE) {
		db = sqlite_open("content.db", 0, NULL);
		db2 = sqlite_open("content_delivered.db", 0, NULL);
		if (errmsg) {
			printf("Error: %s\n", errmsg);
			free (errmsg);
		} else {
			// query all content pieces
			snprintf(query, 4096, "SELECT filename,ID FROM content WHERE valid_to>%lu AND valid_from<%lu",(unsigned long)time(NULL),(unsigned long)time(NULL));
			if (sqlite_compile(db,query,&sqltail,&pVm,&errmsg) == SQLITE_OK) {
				content_pushed=0;
				while (sqlite_step(pVm,&pN,&pazValue,&pazColName) == SQLITE_ROW) {
					
					/*
					 * The device is unknown,
					 * if the content is marked as default
					 * send it, even the device does not support it
					 */

					memset(usemime,'\0',sizeof(usemime));
					memset(usefile,'\0',sizeof(usefile));
					memset(useext,'\0',sizeof(useext));
					
					if (get_content_for_device(d->d,(char*)(pazValue[1]),(char*)(pazValue[0]),(char*)usefile,(char*)usemime,useext) == 0) {
						
						printf("(hci%i): PUSHING to %s: %s (%s - %s).\n",me->dev_id,d->d->name,usefile,usemime,useext);
						if (obex_push(usefile,usemime,useext,me,d)==0) {
							addr_str = my_ba2str(&(d->d->addr));
							if (atoi(d->d->ua_id) == 0) {
								snprintf(query, 4096, "INSERT INTO content_delivered VALUES ('%s', '%s', '%s', '%li')",usefile,"unknown device",addr_str,(unsigned long)time(NULL) );
							} else {
								snprintf(query, 4096, "INSERT INTO content_delivered VALUES ('%s', '%s', '%s', '%li')",usefile,d->d->name,addr_str,(unsigned long)time(NULL) );
							}
							free(addr_str);
							sqlite_exec(db2, query, NULL, NULL, &errmsg);
							if (errmsg) {
								printf("Error: %s\nFile: %s\nLine: %i\n", errmsg, __FILE__,__LINE__);
								free (errmsg);
							}
						}
						
						content_pushed++;
						
						if (d->status == STATUS_REJECTED 	|| 
							d->status == STATUS_FAILED) 
								break;
								
						/*
						 * Give slow devices a few seconds to get back to normal operations
						 */		
						sleep(2);				
					} else {
						/*
						 * There is really no content available
						 */
					}
				}
				
				/*
				 * There was no content for the device,
				 * so put it to state FINISHED
				 */
				if (content_pushed == 0) {
					d->status = STATUS_FINISHED;
					d->tstamp = time(NULL);
					printf("(hci%i): No content was available.\n",me->dev_id);
				}
				
				sqlite_finalize(pVm,NULL);
				pVm = NULL;
				sqltail = NULL;
			} else {
				if (errmsg) {
					printf("Query: %s\n",query);
					printf("Error: %s\n", errmsg);
					free (errmsg);
				}
			}
		}
		sqlite_close(db);
		sqlite_close(db2);
		
//		printf("thread_push(2): Removing addr at p:%p\n",&(d->d->addr));
		remove_locked_device(l,&(d->d->addr));
	} else {
		printf("thread_push: Device was locked meanwhile.\n");
	}
	
	
	if (usemime) { free(usemime); usemime=NULL; }
	if (usefile) { free(usefile); usefile=NULL; }
	if (useext) { free(useext); useext=NULL; }
	
	me->status = STATUS_READY;
	me->tstamp = time(NULL);
	
	free(param);
	
	return NULL;
}
