#ifndef BTCONSTANTS_H_
#define BTCONSTANTS_H_

#include "bluetooth/bluetooth.h"
#include "bluetooth/hci.h"
#include "bluetooth/rfcomm.h"
#include "bluetooth/hci_lib.h"
#include "bluetooth/sdp.h"
#include "bluetooth/sdp_lib.h"

#define TRUE				1
#define	FALSE				0

/*
 * Search for DEVICES_MAX local devices on startup 
 */
#define DEVICES_MAX 		12
/*
 * Use DEVICES_SCANNER devices for scanning purposes 
 */
#define DEVICES_SCANNER		2
/*
 * Use DEVICES_PUSH devices for push purpose 
 */
#define DEVICES_PUSH		10

/*
 * Perform actions all x seconds
 */
#define DEVICE_TICKRATE		2
/*
 * Scan actions should take max. DEVICE_SCANTIME seconds
 */
#define	DEVICE_SCANTIME		12
#define DEVICE_NAME_PREFIX	"biotronik"

/*
 * Default timeout for HCI/SDP ops
 * in milliseconds
 */
#define DEVICE_DEF_TIMEOUT	10000

/*
 * How many retries after TIMEOUT
 */
#define DEVICE_RETRY_TIMOUT 3


/*
 * How many seconds to free a locked device
 * This forced freeing will only be done in emergency 
 * if the device is locked too long.
 */
#define DEVICE_LOCK_TIMEOUT 300

/*
 * Remote device timeout
 * after DEVICE_RTIMEOUT seconds the found device will be removed
 * if it was not found within the last DEVICE_RTIMEOUT seconds
 */
#define DEVICE_RTIMEOUT 	12000

/*
 * Push content to unknown devices
 */
#define PUSH_UNKNOWNDEV		1

#define PURPOSE_SCAN		1
#define	PURPOSE_PUSH		2
#define	STATUS_PROCESSING	1
#define	STATUS_FAILED		2
#define	STATUS_FINISHED		3
#define	STATUS_READY		4
#define	STATUS_REJECTED		5
#define STATUS_UNKNOWNDEV	6
#define STATUS_TIMEOUT 		7

#define SEM_SCAN			0
#define SEM_LOCKS			1
#define SEM_LOCKSLIST		2

#define SDP_SPACER 			0xFF
#define OBEX_PUT_FINAL 		0x82
#define OBEX_PUT_CHUNK 		0x02
#define OBEX_HI_NAME		0x01
#define OBEX_HI_LENGTH		0xC3
#define OBEX_HI_TYPE		0x42
#define OBEX_HI_HTTP		0x47
#define OBEX_HI_ENDOFBODY	0x49
#define OBEX_HI_BODYCHUNK	0x48

typedef struct bp_sdp 			bp_sdp_t;
typedef struct bp_sdp_chunks 	bp_sdp_chunks_t;
typedef struct bp_hci_chunks 	bp_hci_chunks_t;
typedef struct device 			device_t;
typedef struct device_caps 		device_caps_t;
typedef struct mydevices 		mydevices_t;
typedef struct founddevices 	founddevices_t;
typedef struct thread_param 	thread_param_t;
typedef struct lockeddevices 	lockeddevices_t;
typedef struct content 			content_t;

struct content {
	char filename[100];
	char mime1[50];
	char mime2[50];
	char mime3[50];
	int minscreenx;
	int minscreeny;
	int maxscreenx;
	int maxscreeny;
};

struct thread_param {
	void **remote_device_list;
	void *local_device_list;
	void *use_this_device;
	void *use_this_remote_device;
	void **locked_devices;
};

struct mydevices {
	char *name;
    struct hci_dev_info 	di;
	char purpose;
	time_t tstamp;
	pthread_t thread_id;
	char status;
	int dev_id;
	mydevices_t *next;
};

struct lockeddevices {
	time_t tstamp;
	bdaddr_t addr;
	lockeddevices_t *next;
};

struct founddevices {
	char status;
	char retries;
	time_t tstamp;
	device_t *d;
	founddevices_t *next;
};

struct bp_sdp {
 	unsigned int attrId;
	unsigned char datatype;
	void *data;
	unsigned char datalength;
	bp_sdp_t *next;
};

struct bp_sdp_chunks {
	char *chunk[20];
	char *opush;
	char opush_channel;
};

struct bp_hci_chunks {
	char *mac;
	char *lmp;
	char *features;
	char *name;
};

struct search_context {
	char		*svc;		/* Service */
	uuid_t		group;		/* Browse group */
	int			view;		/* View mode */
	uint32_t	handle;		/* Service record handle */
};


struct device_caps {
	char *group;
	device_caps_t *next;
};

struct device {
	char *name;
	bdaddr_t addr;
	unsigned char ua_id[32];
	unsigned int ua_match;
	unsigned int screen_x;
	unsigned int screen_y;
	unsigned int bpp;
	char channel;
	unsigned char is_phone;
	device_caps_t *caps;
};

// A0 0 7 10 0 F A0
struct obex_connect_response {
	unsigned char status;
	unsigned char length[2];
	unsigned char obex_version;
	unsigned char flags;
	unsigned char maxlen[2];
};

#endif /*BTCONSTANTS_H_*/
